from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from desdeo_problem.Problem import MOProblem
from desdeo_tools.interaction.request import BaseRequest, SimplePlotRequest
from desdeo_tools.scalarization.ASF import AugmentedGuessASF, MaxOfTwoASF, PointMethodASF, SimpleASF, StomASF
from desdeo_tools.scalarization.Scalarizer import Scalarizer
from desdeo_tools.solver.ScalarSolver import ScalarMethod, ScalarMinimizer

from desdeo_mcdm.interactive.InteractiveMethod import InteractiveMethod
from desdeo_mcdm.utilities.solvers import payoff_table_method


class NimbusException(Exception):
    """Risen when an error related to NIMBUS is encountered.

    """

    pass


class NimbusClassificationRequest(BaseRequest):
    """A request to handle the classification of objectives in the synchronous NIMBUS method.

    Args:
        method (NIMBUS): The instance of the NIMBUS method the request should be initialized for.
        ref (np.ndarray): Objective values used as a reference the decision maker is classifying the objectives.

    Attributes:
        self._valid_classifications (List[str]): The valid classifications. Defaults is ['<', '<=', '=', '>=', '0']
    """

    def __init__(self, method: NIMBUS, ref: np.ndarray):
        msg = (
            "Please classify each of the objective values in one of the following categories:"
            "\n\t1. values should improve '<'"
            "\n\t2. values should improve until some desired aspiration level is reached '<='"
            "\n\t3. values with an acceptable level '='"
            "\n\t4. values which may be impaired until some upper bound is reached '>='"
            "\n\t5. values which are free to change '0'"
            "\nProvide the aspiration levels and upper bounds as a vector. For categories 1, 3, and 5,"
            "the value in the vector at the objective's position is ignored. Suppy also the number of maximum"
            "solutions to be generated."
        )

        self._method = method
        self._valid_classifications = ["<", "<=", "=", ">=", "0"]
        content = {
            "message": msg,
            "objective_values": ref,
            "classifications": [None],
            "levels": [None],
            "number_of_solutions": 1,
        }
        super().__init__("classification_preference", "required", content=content)

    def validator(self, response: Dict) -> None:
        """Validates a dictionary containing the response of a decision maker. Should contain the keys 
        'classifications', 'levels', and 'number_of_solutions'.

        'classifications' should be a list of strings, where the number of
        elements is equal to the number of objectives being classified, and
        the elements are found in `_valid_classifications`. 'levels' should
        have either aspiration levels or bounds for each objective depending
        on that objective's classification. 'number_of_solutions' should be
        an integer between 1 and 4 indicating the number of intermediate solutions to be
        computed.

        Args:
            response (Dict): See the documentation for `validator`.

        Raises:
            NimbusException: Some discrepancy is encountered in the parsing of the response.
        """
        if "classifications" not in response:
            raise NimbusException("'classifications' entry missing.")

        if "levels" not in response:
            raise NimbusException("'levels' entry missing.")

        if "number_of_solutions" not in response:
            raise NimbusException("'number_of_solutions' entry missing.")

        # check the classifications
        is_valid_cls = map(lambda x: x in self._valid_classifications, response["classifications"],)

        if not all(list(is_valid_cls)):
            raise NimbusException(f"Invalid classificaiton found in {response['classifications']}")

        # check the levels
        if len(np.array(response["levels"]).squeeze()) != self._method._problem.n_of_objectives:
            raise NimbusException(f"Wrong number of levels supplied in {response['levels']}")

        improve_until_inds = np.where(np.array(response["classifications"]) == "<=")[0]

        impaire_until_inds = np.where(np.array(response["classifications"]) == ">=")[0]

        if len(improve_until_inds) > 0:
            # some objectives classified to be improved until some level
            if not np.all(
                np.array(response["levels"])[improve_until_inds] >= self._method._ideal[improve_until_inds]
            ) or not np.all(
                np.array(response["levels"])[improve_until_inds] <= self._method._nadir[improve_until_inds]
            ):
                raise NimbusException("Given levels must be between the nadir and ideal points!")

        if len(impaire_until_inds) > 0:
            # some objectives classified to be improved until some level
            if not np.all(
                np.array(response["levels"])[impaire_until_inds] >= self._method._ideal[impaire_until_inds]
            ) or not np.all(
                np.array(response["levels"])[impaire_until_inds] <= self._method._nadir[impaire_until_inds]
            ):
                raise NimbusException("Given levels must be between the nadir and ideal points!")

        # check maximum number of solutions
        if response["number_of_solutions"] > 4 or response["number_of_solutions"] < 1:
            raise NimbusException("The number of solutions must be between 1 and 4.")

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class NimbusSaveRequest(BaseRequest):
    """A request to handle archiving of the solutions computed with NIMBUS.

    Args:
        solution_vectors (List[np.ndarray]): A list of numpy arrays each representing a decision variable vector.
        objective_vectors (List[np.ndarray]): A list of numpy arrays each representing an objective vector.

    Note:
        The objective vector at position 'i' in `objective_vectors` should correspond to the decision variables at
        position 'i' in `solution_vectors`.
    """

    def __init__(
        self, solution_vectors: List[np.ndarray], objective_vectors: List[np.ndarray],
    ):
        msg = (
            "Please specify which solutions shown you would like to save for later viewing. Supply the "
            "indices of such solutions as a list, or supply an empty list if none of the shown soulutions "
            "should be saved."
        )
        content = {
            "message": msg,
            "solutions": solution_vectors,
            "objectives": objective_vectors,
            "indices": [],
        }
        super().__init__("classification_preference", "required", content=content)

    def validator(self, response: Dict) -> None:
        """Validates a response dictionary. The dictionary should contain the keys 'indices'.

        'indices' should be a list of integers representing an index to the
        lists `solutions_vectors` and `objective_vectors`.

        Args:
            response (Dict): See the documentation for `validator`.

        Raises:
            NimbusException: Some discrepancy is encountered in the parsing of `response`.
        """
        if "indices" not in response:
            raise NimbusException("'indices' entry missing")

        if not response["indices"]:
            # nothing to save, continue to next state
            return

        if len(response["indices"]) > len(self.content["objectives"]) or np.min(response["indices"]) < 0:
            # wrong number of indices
            raise NimbusException(f"Invalid indices {response['indices']}")

        if np.max(response["indices"]) >= len(self.content["objectives"]) or np.min(response["indices"]) < 0:
            # out of bounds index
            raise NimbusException(f"Incides {response['indices']} out of bounds.")

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class NimbusIntermediateSolutionsRequest(BaseRequest):
    """A request to handle the computation of intermediate points between two previously computed points.

    Args:
        solution_vectors (List[np.ndarray]): A list of numpy arrays each representing a decision variable vector.
        objective_vectors (List[np.ndarray]): A list of numpy arrays each representing an objective vector.

    Note:
        The objective vector at position 'i' in `objective_vectors` should correspond to the decision variables at
        position 'i' in `solution_vectors`. Only the two first entries in each of the lists is relevant. The
        rest is ignored.
    """

    def __init__(
        self, solution_vectors: List[np.ndarray], objective_vectors: List[np.ndarray],
    ):
        msg = (
            "Would you like to see intermediate solutions between two previusly computed solutions? "
            "If so, please supply two indices corresponding to the solutions."
        )

        content = {
            "message": msg,
            "solutions": solution_vectors,
            "objectives": objective_vectors,
            "indices": [],
            "number_of_desired_solutions": 0,
        }

        super().__init__("classification_preference", "required", content=content)

    def validator(self, response: Dict):
        """Validates a response dictionary. The dictionary should contain the keys 'indices' and 'number_of_solutions'.

        'indices' should be a list of integers representing an index to the
        lists `solutions_vectors` and `objective_vectors`. 'number_of_solutions' should be an integer greater or equal 
        to 1.

        Args:
            response (Dict): See the documentation for `validator`.

        Raises:
            NimbusException: Some discrepancy is encountered in the parsing of `response`.
        """
        if "indices" not in response:
            raise NimbusException("'indices' entry missing.")

        if "number_of_desired_solutions" not in response:
            raise NimbusException("'number_of_desired_solutions' entry missing.")

        if response["number_of_desired_solutions"] < 0:
            raise NimbusException(f"Invalid number of desired solutions {response['number_of_desired_solutions']}.")

        if not response["indices"] and response["number_of_desired_solutions"] > 0:
            raise NimbusException("Indices supplied yet number of desired soutions is greater than zero.")

        if response["indices"] and response["number_of_desired_solutions"] == 0:
            raise NimbusException("Indices not supplied yet number of desired soutions is zero.")

        if not response["indices"]:
            return

        if np.max(response["indices"]) >= len(self.content["objectives"]) or np.min(response["indices"]) < 0:
            # indices out of bounds
            raise NimbusException(f"Invalid indices {response['indices']}")

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class NimbusMostPreferredRequest(BaseRequest):
    """A request to handle the indication of a preferred point.

    Args:
        solution_vectors (List[np.ndarray]): A list of numpy arrays each representing a decision variable vector.
        objective_vectors (List[np.ndarray]): A list of numpy arrays each representing an objective vector.

    Note:
        The objective vector at position 'i' in `objective_vectors` should correspond to the decision variables at
        position 'i' in `solution_vectors`. Only the two first entries in each of the lists is relevant. The preferred
        solution will be selected from `objective_vectors`.

    """

    def __init__(
        self, solution_vectors: List[np.ndarray], objective_vectors: List[np.ndarray],
    ):
        msg = "Please select your most preferred solution and whether you would like to continue. "

        content = {
            "message": msg,
            "solutions": solution_vectors,
            "objectives": objective_vectors,
            "index": -1,
            "continue": True,
        }

        super().__init__("classification_preference", "required", content=content)

    def validator(self, response: Dict):
        """Validates a response dictionary. The dictionary should contain the keys 'index' and 'continue'.

        'index' is an integer and should indicate the index of the preferred solution is `objective_vectors`. 
        'continue' is a boolean and indicates whether to stop or continue the iteration of Synchronous NIMBUS.

        Args:
            response (Dict): See the documentation for `validator`.

        Raises:
            NimbusException: Some discrepancy is encountered in the parsing of `response`.
        """
        if "index" not in response:
            raise NimbusException(f"'index' entry missing.")

        if "continue" not in response:
            raise NimbusException(f"'continue' entry missing.")

        if not type(response["index"]) == int:
            raise NimbusException(f"The index must be a single integer, found {response['index']}")

        if not type(response["continue"]) == bool:
            raise NimbusException(f"Continue must be a boolean value, found {response['index']}")

        if not (response["index"] >= 0 and response["index"] < len(self.content["objectives"])):
            raise NimbusException(
                (
                    f"The index must be a positive integer less than "
                    f"{len(self.content['objectives'])}, found {response['index']}."
                )
            )

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class NimbusStopRequest(BaseRequest):
    """A request to handle the termination of Synchronous NIMBUS.

    Args:
        solutions_final (np.ndarray): A numpy array containing the final decision variable values.
        objective_final (np.ndarray): A numpy array containing the final objective variables which correspond to
        `solution_final`.

    Note:
        This request expects no response.
    """

    def __init__(
        self, solution_final: np.ndarray, objective_final: np.ndarray,
    ):
        msg = "The final solution computed."

        content = {
            "message": msg,
            "solution": solution_final,
            "objective": objective_final,
        }

        super().__init__("classification_preference", "no_interaction", content=content)


class NIMBUS(InteractiveMethod):
    """Implements the synchronous NIMBUS algorithm.

    Args:
        problem (MOProblem): The problem to be solved.
        scalar_method (Optional[ScalarMethod], optional): The method used to solve
            the various ASF minimization problems present in the method. Defaults to None.
    """

    def __init__(self, problem: MOProblem, scalar_method: Optional[ScalarMethod] = None):
        # check if ideal and nadir are defined
        if problem.ideal is None or problem.nadir is None:
            # TODO: use same method as defined in scalar_method
            ideal, nadir = payoff_table_method(problem)
            self._ideal = ideal
            self._nadir = nadir
        else:
            self._ideal = problem.ideal
            self._nadir = problem.nadir

        self._scalar_method = scalar_method

        # generate Pareto optimal starting point
        asf = SimpleASF(np.ones(self._ideal.shape))
        scalarizer = Scalarizer(
            lambda x: problem.evaluate(x).objectives,
            asf,
            scalarizer_args={"reference_point": np.atleast_2d(self._ideal)},
        )

        if problem.n_of_constraints > 0:
            _con_eval = lambda x: problem.evaluate(x).constraints.squeeze()
        else:
            _con_eval = None

        solver = ScalarMinimizer(
            scalarizer, problem.get_variable_bounds(), constraint_evaluator=_con_eval, method=self._scalar_method,
        )
        # TODO: fix tools to check for scipy methods in general and delete me!
        solver._use_scipy = True

        res = solver.minimize(problem.get_variable_upper_bounds() / 2)

        if res["success"]:
            self._current_solution = res["x"]
            self._current_objectives = problem.evaluate(self._current_solution).objectives.squeeze()

        self._archive_solutions = []
        self._archive_objectives = []
        self._state = "classify"

        super().__init__(problem)

    def start(self) -> Tuple[NimbusClassificationRequest, SimplePlotRequest]:
        """Return the first request to start iterating NIMBUS.
        
        Returns:
            Tuple[NimbusClassificationRequest, SimplePlotRequest]: The first request and
            and a plot request to visualize relevant data.
        """
        return self.request_classification()

    def request_classification(self,) -> Tuple[NimbusClassificationRequest, SimplePlotRequest]:
        return (
            NimbusClassificationRequest(self, self._current_objectives.squeeze()),
            None,
        )

    def create_plot_request(self, objectives: np.ndarray, msg: str) -> SimplePlotRequest:
        """Used to create a plot request for visualizing objective values.

        Args:
            objectives (np.ndarray): A 2D numpy array containing objective vectors to be visualized.
            msg (str): A message to be displayed in the context of a visualization.

        Returns:
            SimplePlotRequest: A plot request to create a visualization.
        """
        dimensions_data = pd.DataFrame(
            index=["minimize", "ideal", "nadir"], columns=self._problem.get_objective_names(),
        )
        dimensions_data.loc["minimize"] = self._problem._max_multiplier
        dimensions_data.loc["ideal"] = self._ideal
        dimensions_data.loc["nadir"] = self._nadir

        data = pd.DataFrame(objectives, columns=self._problem.get_objective_names())

        plot_request = SimplePlotRequest(data=data, dimensions_data=dimensions_data, message=msg,)

        return plot_request

    def handle_classification_request(
        self, request: NimbusClassificationRequest
    ) -> Tuple[NimbusSaveRequest, SimplePlotRequest]:
        """Handles a classification request.
        
        Args:
            request (NimbusClassificationReuest): A classification request with the
                response attribute set.
        
        Returns:
            Tuple[NimbusSaveRequest, SimplePlotRequest]: A NIMBUS save request and a plot request
            with the solutions the decision maker can choose from to save for alter use.
        """
        improve_inds = np.where(np.array(request.response["classifications"]) == "<")[0]

        acceptable_inds = np.where(np.array(request.response["classifications"]) == "=")[0]

        free_inds = np.where(np.array(request.response["classifications"]) == "0")[0]

        improve_until_inds = np.where(np.array(request.response["classifications"]) == "<=")[0]

        impaire_until_inds = np.where(np.array(request.response["classifications"]) == ">=")[0]

        # calculate the new solutions
        return self.calculate_new_solutions(
            int(request.response["number_of_solutions"]),
            np.array(request.response["levels"]),
            improve_inds,
            improve_until_inds,
            acceptable_inds,
            impaire_until_inds,
            free_inds,
        )

    def handle_save_request(
        self, request: NimbusSaveRequest
    ) -> Tuple[NimbusIntermediateSolutionsRequest, SimplePlotRequest]:
        """Handles a save request.
        
        Args:
            request (NimbusSaveRequest): A save request with the response attribute set.
        
        Returns:
            Tuple[NimbusIntermediateSolutionsRequest, SimplePlotRequest]: Return an
            intermediate solution request where the decision maker can specify whether they
            would like to see intermediate solution between two previously computed solutions.
            The plot request has the available solutions.
        """
        return self.save_solutions_to_archive(
            np.array(request.content["objectives"]),
            np.array(request.content["solutions"]),
            np.array(request.response["indices"]),
        )

    def handle_intermediate_solutions_request(
        self, request: NimbusIntermediateSolutionsRequest
    ) -> Tuple[
        Union[NimbusSaveRequest, NimbusMostPreferredRequest], SimplePlotRequest,
    ]:
        """Handles an intermediate solutions request.
        
        Args:
            request (NimbusIntermediateSolutionsRequest): A NIMBUS intermediate solutions
                request with the response attribute set.
        
        Returns:
            Tuple[Union[NimbusSaveRequest, NimbusMostPreferredRequest], SimplePlotRequest,]:
            Return either a save request or a preferred solution request. The former is returned if the
            decision maker wishes to see intermediate points, the latter otherwise. Also a plot request is
            returned with the solutions available in it.
        """
        if request.response["indices"]:
            return self.compute_intermediate_solutions(
                np.array(request.content["solutions"])[request.response["indices"]],
                int(request.response["number_of_desired_solutions"]),
            )

        return self.request_most_preferred_solution(
            np.array(request.content["solutions"]), np.array(request.content["objectives"]),
        )

    def handle_most_preferred_request(
        self, request: NimbusMostPreferredRequest
    ) -> Tuple[Union[NimbusClassificationRequest, NimbusStopRequest], SimplePlotRequest]:
        """Handles a preferres solution request.
        
        Args:
            request (NimbusMostPreferredRequest): A NIMBUS preferred solution request with the
                response attribute set.
        
        Returns:
            Tuple[Union[NimbusClassificationRequest, NimbusStopRequest], SimplePlotRequest]:
            Return a classificaiton request if the decision maker wishes to continue. If the
            decision maker wishes to stop, return a stop request. Also return a plot 
            request with all the solutions saved so far.
        """
        self.update_current_solution(
            np.array(request.content["solutions"]),
            np.array(request.content["objectives"]),
            np.array(request.response["index"]),
        )

        if not request.response["continue"]:
            return self.request_stop()

        else:
            return self.request_classification()

    def request_stop(self) -> Tuple[NimbusStopRequest, SimplePlotRequest]:
        """Create a NimbusStopRequest based on self.
        
        Returns:
            Tuple[NimbusStopRequest, SimplePlotRequest]: A stop request and a plot
            request with the final solution chosen in it.
        """
        request = NimbusStopRequest(self._current_solution, self._current_objectives)

        msg = "Final solution reached"
        plot_request = self.create_plot_request(np.atleast_2d(self._current_objectives), msg)

        return request, plot_request

    def request_most_preferred_solution(
        self, solutions: np.ndarray, objectives: np.ndarray
    ) -> Tuple[NimbusMostPreferredRequest, SimplePlotRequest]:
        """Create a NimbusMostPreferredRequest.

        Args:
            solutions (np.ndarray): A 2D numpy array of decision variable vectors.
            objectives (np.ndarray): A 2D numpy array of objective value vectors.

        Returns:
            Tuple[NimbusMostPreferredRequest, SimplePlotRequest]: The requests based on the given arguments.

        Note:
            The 'i'th decision variable vector in `solutions` should correspond to the 'i'th objective value vector in
            `objectives`.

        """
        # request most preferred solution
        request = NimbusMostPreferredRequest(list(solutions), list(objectives))

        msg = "Computed solutions"
        plot_request = self.create_plot_request(objectives, msg)

        return request, plot_request

    def compute_intermediate_solutions(
        self, solutions: np.ndarray, n_desired: int,
    ) -> Tuple[NimbusSaveRequest, SimplePlotRequest]:
        """Computs intermediate solution between two solutions computed earlier.

        Args:
            solutions (np.ndarray): The solutions between which the intermediat solutions should
                be computed.
            n_desired (int): The number of intermediate solutions desired.
        
        Raises:
            NimbusException
        
        Returns:
            Tuple[NimbusSaveRequest, SimplePlotRequest]: A save request with the compured intermediate
            points, and a plot request to visualize said points.
        """
        # vector between the two solutions
        between = solutions[0] - solutions[1]
        norm = np.linalg.norm(between)
        between_norm = between / norm

        # the plus 2 assumes we are interested only in n_desired points BETWEEN the
        # two supplied solutions
        step_size = norm / (2 + n_desired)

        intermediate_points = np.array([solutions[1] + i * step_size * between_norm for i in range(1, n_desired + 1)])

        # project each of the intermediate solutions to the Pareto front
        intermediate_solutions = np.zeros(intermediate_points.shape)
        intermediate_objectives = np.zeros((n_desired, self._problem.n_of_objectives))
        asf = PointMethodASF(self._nadir, self._ideal)

        for i in range(n_desired):
            scalarizer = Scalarizer(
                lambda x: self._problem.evaluate(x).objectives,
                asf,
                scalarizer_args={"reference_point": self._problem.evaluate(intermediate_points[i]).objectives},
            )

            if self._problem.n_of_constraints > 0:
                cons = lambda x: self._problem.evaluate(x).constraints.squeeze()
            else:
                cons = None

            solver = ScalarMinimizer(scalarizer, self._problem.get_variable_bounds(), cons, method=self._scalar_method,)

            res = solver.minimize(self._current_solution)
            intermediate_solutions[i] = res["x"]
            intermediate_objectives[i] = self._problem.evaluate(res["x"]).objectives

        # create appropiate requests
        save_request = NimbusSaveRequest(list(intermediate_solutions), list(intermediate_objectives))

        msg = "Computed intermediate solutions"
        plot_request = self.create_plot_request(intermediate_objectives, msg)

        return save_request, plot_request

    def save_solutions_to_archive(
        self, objectives: np.ndarray, decision_variables: np.ndarray, indices: List[int],
    ) -> Tuple[NimbusIntermediateSolutionsRequest, None]:
        """Save solutions to the archive. Saves also the corresponding objective function
        values.
        
        Args:
            objectives (np.ndarray): Available objectives.
            decision_variables (np.ndarray): Available solutions.
            indices (List[int]): Indices of the solutions to be saved.
        
        Returns:
            Tuple[NimbusIntermediateSolutionsRequest, None]: An intermediate solutins request asking the
            decision maker whether they would like to generate intermediata solutions between two existing solutions.
            Also returns a plot request to visualize the available solutions between which the intermediate solutions
            should be computed.
        """
        self._archive_objectives.extend(list(objectives[indices]))
        self._archive_solutions.extend(list(decision_variables[indices]))

        mask = np.ones(objectives.shape[0], dtype=bool)
        mask[indices] = False

        req_objectives = self._archive_objectives + list(objectives[mask])
        req_solutions = self._archive_solutions + list(decision_variables[mask])

        # create intermediate solutions request
        request = NimbusIntermediateSolutionsRequest(req_solutions, req_objectives)

        msg = "Computed new solutions"
        plot_request = self.create_plot_request(req_objectives, msg)

        return request, plot_request

    def calculate_new_solutions(
        self,
        number_of_solutions: int,
        levels: np.ndarray,
        improve_inds: np.ndarray,
        improve_until_inds: np.ndarray,
        acceptable_inds: np.ndarray,
        impaire_until_inds: np.ndarray,
        free_inds: np.ndarray,
    ) -> Tuple[NimbusSaveRequest, SimplePlotRequest]:
        """Calcualtes new solutions based on classifications supplied by the decision maker by
            solving ASF problems.
        
        Args:
            number_of_solutions (int): Number of solutions, should be between 1 and 4.
            levels (np.ndarray): Aspiration and upper bounds relevant to the some of the classifications.
            improve_inds (np.ndarray): Indices corresponding to the objectives which should be improved.
            improve_until_inds (np.ndarray): Like above, but improved until an aspiration level is reached.
            acceptable_inds (np.ndarray): Indices of objectives which are acceptable as they are now.
            impaire_until_inds (np.ndarray): Indices of objectives which may be impaired until an upper limit is
                reached.
            free_inds (np.ndarray): Indices of objectives which may change freely.
        
        Returns:
            Tuple[NimbusSaveRequest, SimplePlotRequest]: A save request with the newly computed soutions, and 
            a plot request to visualize said solutions.
        """
        results = []

        # always computed
        asf_1 = MaxOfTwoASF(self._nadir, self._ideal, improve_inds, improve_until_inds)

        def cons_1(
            x: np.ndarray,
            f_current: np.ndarray = self._current_objectives,
            levels: np.ndarray = levels,
            improve_until_inds: np.ndarray = improve_until_inds,
            improve_inds: np.ndarray = improve_inds,
            impaire_until_inds: np.ndarray = impaire_until_inds,
        ):
            f = self._problem.evaluate(x).objectives.squeeze()

            res_1 = f_current[improve_inds] - f[improve_inds]
            res_2 = f_current[improve_until_inds] - f[improve_until_inds]
            res_3 = levels[impaire_until_inds] - f_current[impaire_until_inds]

            res = np.hstack((res_1, res_2, res_3))

            if self._problem.n_of_constraints > 0:
                res_prob = self._problem.evaluate(x).constraints.squeeze()

                return np.hstack((res_prob, res))

            else:
                return res

        scalarizer_1 = Scalarizer(
            lambda x: self._problem.evaluate(x).objectives, asf_1, scalarizer_args={"reference_point": levels},
        )

        solver_1 = ScalarMinimizer(
            scalarizer_1, self._problem.get_variable_bounds(), cons_1, method=self._scalar_method,
        )

        res_1 = solver_1.minimize(self._current_solution)
        results.append(res_1)

        if number_of_solutions > 1:
            # create the reference point needed in the rest of the ASFs
            z_bar = np.zeros(self._problem.n_of_objectives)
            z_bar[improve_inds] = self._ideal[improve_inds]
            z_bar[improve_until_inds] = levels[improve_until_inds]
            z_bar[acceptable_inds] = self._current_objectives[acceptable_inds]
            z_bar[impaire_until_inds] = levels[impaire_until_inds]
            z_bar[free_inds] = self._nadir[free_inds]

            # second ASF
            asf_2 = StomASF(self._ideal)

            # cons_2 can be used in the rest of the ASF scalarizations, it's not a bug!
            if self._problem.n_of_constraints > 0:
                cons_2 = lambda x: self._problem.evaluate(x).constraints.squeeze()
            else:
                cons_2 = None

            scalarizer_2 = Scalarizer(
                lambda x: self._problem.evaluate(x).objectives, asf_2, scalarizer_args={"reference_point": z_bar},
            )

            solver_2 = ScalarMinimizer(
                scalarizer_2, self._problem.get_variable_bounds(), cons_2, method=self._scalar_method,
            )

            res_2 = solver_2.minimize(self._current_solution)
            results.append(res_2)

        if number_of_solutions > 2:
            # asf 3
            asf_3 = PointMethodASF(self._nadir, self._ideal)

            scalarizer_3 = Scalarizer(
                lambda x: self._problem.evaluate(x).objectives, asf_3, scalarizer_args={"reference_point": z_bar},
            )

            solver_3 = ScalarMinimizer(
                scalarizer_3, self._problem.get_variable_bounds(), cons_2, method=self._scalar_method,
            )

            res_3 = solver_3.minimize(self._current_solution)
            results.append(res_3)

        if number_of_solutions > 3:
            # asf 4
            asf_4 = AugmentedGuessASF(self._nadir, self._ideal, free_inds)

            scalarizer_4 = Scalarizer(
                lambda x: self._problem.evaluate(x).objectives, asf_4, scalarizer_args={"reference_point": z_bar},
            )

            solver_4 = ScalarMinimizer(
                scalarizer_4, self._problem.get_variable_bounds(), cons_2, method=self._scalar_method,
            )

            res_4 = solver_4.minimize(self._current_solution)
            results.append(res_4)

        # create the save request
        solutions = [res["x"] for res in results]
        objectives = [self._problem.evaluate(x).objectives.squeeze() for x in solutions]

        save_request = NimbusSaveRequest(solutions, objectives)

        msg = "Computed new solutions."
        plot_request = self.create_plot_request(objectives, msg)

        return save_request, plot_request

    def update_current_solution(self, solutions: np.ndarray, objectives: np.ndarray, index: int) -> None:
        """Update the state of self with a new current solution and the corresponding objective values. This solution is
        used in the classification phase of synchronous NIMBUS.

        Args:
            solutions (np.ndarray): A 2D numpy array of decision variable vectors.
            objectives (np.ndarray): A 2D numpy array of objective value vectors.
            index (int): The index of the solution in `solutions` and `objectives`.

        Returns:
            Tuple[NimbusMostPreferredRequest, SimplePlotRequest]: The requests based on the given arguments.

        Note:
            The 'i'th decision variable vector in `solutions` should correspond to the 'i'th objective value vector in
            `objectives`.

        """
        self._current_solution = solutions[index].squeeze()
        self._current_objectives = objectives[index].squeeze()

        return None

    def iterate(
        self,
        request: Union[
            NimbusClassificationRequest,
            NimbusSaveRequest,
            NimbusIntermediateSolutionsRequest,
            NimbusMostPreferredRequest,
            NimbusStopRequest,
        ],
    ) -> Tuple[
        Union[NimbusClassificationRequest, NimbusSaveRequest, NimbusIntermediateSolutionsRequest,],
        Union[SimplePlotRequest, None],
    ]:
        """Implements a finite state machine to iterate over the different steps defined in Synchronous NIMBUS based on a supllied request.
        
        Args:
            request (Union[NimbusClassificationRequest,NimbusSaveRequest,NimbusIntermediateSolutionsRequest,NimbusMostPreferredRequest,NimbusStopRequest,]):
                A request based on the next step in the NIMBUS algorithm is taken.
        
        Raises:
            NimbusException: If a wrong type of request is supllied based on the current state NIMBUS is in.
        
        Returns:
            Tuple[Union[NimbusClassificationRequest,NimbusSaveRequest,NimbusIntermediateSolutionsRequest,],Union[SimplePlotRequest, None],]:
            The next logically sound request.
        """
        if self._state == "classify":
            if type(request) != NimbusClassificationRequest:
                raise NimbusException(
                    f"Expected request type {type(NimbusClassificationRequest)}, was {type(request)}."
                )

            requests = self.handle_classification_request(request)
            self._state = "archive"
            return requests

        elif self._state == "archive":
            if type(request) != NimbusSaveRequest:
                raise NimbusException(f"Expected request type {type(NimbusSaveRequest)}, was {type(request)}.")

            requests = self.handle_save_request(request)
            self._state = "intermediate"
            return requests

        elif self._state == "intermediate":
            if type(request) != NimbusIntermediateSolutionsRequest:
                raise NimbusException(
                    f"Expected request type {type(NimbusIntermediateSolutionsRequest)}, was {type(request)}."
                )

            requests = self.handle_intermediate_solutions_request(request)

            if type(requests[0]) == NimbusSaveRequest:
                self._state = "archive"
            elif type(requests[0]) == NimbusMostPreferredRequest:
                self._state = "preferred"

            return requests

        elif self._state == "preferred":
            if type(request) != NimbusMostPreferredRequest:
                raise NimbusException(f"Expected request type {type(NimbusMostPreferredRequest)}, was {type(request)}.")

            requests = self.handle_most_preferred_request(request)

            if type(requests[0]) == NimbusStopRequest:
                self._state = "end"

            elif type(requests[0]) == NimbusClassificationRequest:
                self._state = "classify"

            return requests

        elif self._state == "end":
            # end
            return request, None

        else:
            # unknown state error
            raise NimbusException(f"Unknown state '{self._state}' encountered.")


if __name__ == "__main__":
    from desdeo_problem.Objective import _ScalarObjective
    from desdeo_problem.Variable import variable_builder
    from desdeo_problem.Constraint import ScalarConstraint

    # create the problem
    def f_1(x):
        res = 4.07 + 2.27 * x[:, 0]
        return -res

    def f_2(x):
        res = 2.60 + 0.03 * x[:, 0] + 0.02 * x[:, 1] + 0.01 / (1.39 - x[:, 0] ** 2) + 0.30 / (1.39 - x[:, 1] ** 2)
        return -res

    def f_3(x):
        res = 8.21 - 0.71 / (1.09 - x[:, 0] ** 2)
        return -res

    def f_4(x):
        res = 0.96 - 0.96 / (1.09 - x[:, 1] ** 2)
        return -res

    def f_5(x):
        return np.max([np.abs(x[:, 0] - 0.65), np.abs(x[:, 1] - 0.65)], axis=0)

    def c_1(x, f=None):
        x = x.squeeze()
        return (x[0] + x[1]) - 0.5

    f1 = _ScalarObjective(name="f1", evaluator=f_1)
    f2 = _ScalarObjective(name="f2", evaluator=f_2)
    f3 = _ScalarObjective(name="f3", evaluator=f_3)
    f4 = _ScalarObjective(name="f4", evaluator=f_4)
    f5 = _ScalarObjective(name="f5", evaluator=f_5)
    varsl = variable_builder(
        ["x_1", "x_2"], initial_values=[0.5, 0.5], lower_bounds=[0.3, 0.3], upper_bounds=[1.0, 1.0],
    )
    c1 = ScalarConstraint("c1", 2, 5, evaluator=c_1)
    problem = MOProblem(variables=varsl, objectives=[f1, f2, f3, f4, f5], constraints=[c1])

    method = NIMBUS(problem, scalar_method="scipy_de")
    reqs = method.request_classification()[0]

    response = {}
    response["classifications"] = ["<", "<=", "=", ">=", "0"]
    response["levels"] = [-6, -3, -5, 8, 0.349]
    response["number_of_solutions"] = 3
    reqs.response = response
    res_1 = method.iterate(reqs)[0]
    res_1.response = {"indices": [0, 1, 2]}

    res_2 = method.iterate(res_1)[0]
    response = {}
    response["indices"] = []
    response["number_of_desired_solutions"] = 0
    res_2.response = response

    res_3 = method.iterate(res_2)[0]
    response_pref = {}
    response_pref["index"] = 1
    response_pref["continue"] = True
    res_3.response = response_pref

    res_4 = method.iterate(res_3)



"""This module contains various utilities used in different interactive methods.

"""

__all__ = [
    "payoff_table_method",
    "payoff_table_method_general",
    "solve_pareto_front_representation",
    "solve_pareto_front_representation_general",
    "weighted_scalarizer",
]


from desdeo_mcdm.utilities.solvers import (
    payoff_table_method,
    payoff_table_method_general,
    solve_pareto_front_representation,
    solve_pareto_front_representation_general,
    weighted_scalarizer,
)



"""Implements various useful solvers.

"""
import logging
from typing import Callable, Optional, Tuple, Union

import numpy as np
from desdeo_problem.Problem import MOProblem

from desdeo_tools.scalarization.ASF import ASFBase, PointMethodASF
from desdeo_tools.scalarization.Scalarizer import Scalarizer
from desdeo_tools.solver.ScalarSolver import ScalarMethod, ScalarMinimizer


class MCDMUtilityException(Exception):
    """Raised when an exception is encountered in some of the utilities.

    """

    pass


def weighted_scalarizer(xs: np.ndarray, ws: np.ndarray) -> np.ndarray:
    """A simple linear weight based scalarizer.
    
    Args:
        xs (np.ndarray): Values to be scalarized.
        ws (np.ndarray): Weights to multiply each value in the summation of xs.
    
    Returns:
        np.ndarray: An array of scalar values with length equal to the first dimension of xs.
    """
    return np.sum(np.atleast_2d(xs) * ws, axis=1)


def payoff_table_method_general(
    objective_evaluator: Callable[[np.ndarray], np.ndarray],
    n_of_objectives: int,
    variable_bounds: np.ndarray,
    constraint_evaluator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    initial_guess: Optional[np.ndarray] = None,
    solver_method: Optional[Union[ScalarMethod, str]] = "scipy_de",
) -> Tuple[np.ndarray, np.ndarray]:
    """Solves a representation for the nadir and ideal points for a
    multiobjective minimization problem with objectives defined as the result
    of some objective evaluator.
    
    Args:
        objective_evaluator (Callable[[np.ndarray], np.ndarray]): The
            evaluator which returns the objective values given a set of
            variabels.
        n_of_objectives (int): Number of objectives returned by calling
            objective_evaluator.
        variable_bounds (np.ndarray): The lower and upper bounds of the
            variables passed as argument to objective_evaluator. Should be a 2D
            numpy array with the limits for each variable being on each row. The
            first column should contain the lower bounds, and the second column
            the upper bounds. Use np.inf to indicate no bounds.
        constraint_evaluator (Optional[Callable[[np.ndarray], np.ndarray]], optional):
            An evaluator accepting the same arguments as
            objective_evaluator, which returns the constraint values of the
            multiobjective minimization problem being solved. A negative
            constraint value indicates a broken constraint. Defaults to None.
        initial_guess (Optional[np.ndarray], optional): The initial guess
            used for the variable values while solving the payoff table. The
            relevancy of this parameter depends on the solver_method being used.
            Defaults to None.
        solver_method (Optional[Union[ScalarMethod, str]], optional): The
            method to solve the scalarized problems in the payoff table method.
            Defaults to "scipy_de", which ignores initial_guess.
    
    Returns:
        Tuple[np.ndarray, np.ndarray]: The representations computed using the
        payoff table for the ideal and nadir points respectively.
    """
    scalarizer = Scalarizer(objective_evaluator, weighted_scalarizer, scalarizer_args={"ws": None},)

    solver = ScalarMinimizer(scalarizer, variable_bounds, constraint_evaluator, solver_method,)

    ws = np.eye(n_of_objectives)
    po_table = np.zeros((n_of_objectives, n_of_objectives))
    if initial_guess is None:
        initial_guess = variable_bounds[:, 0]

    for i in range(n_of_objectives):
        scalarizer._scalarizer_args = {"ws": ws[i]}
        opt_res = solver.minimize(initial_guess)
        if not opt_res["success"]:
            print("Unsuccessful optimization result encountered while computing a payoff table!")
        po_table[i] = objective_evaluator(opt_res["x"])

    ideal = np.diag(po_table)
    nadir = np.max(po_table, axis=0)

    return ideal, nadir


def payoff_table_method(
    problem: MOProblem,
    initial_guess: Optional[np.ndarray] = None,
    solver_method: Optional[Union[ScalarMethod, str]] = "scipy_de",
) -> Tuple[np.ndarray, np.ndarray]:
    """Uses the payoff table method to solve for the ideal and nadir points of a MOProblem.
    Call through to payoff_table_method_general.
    
    Args:
        problem (MOProblem): The problem defined as a MOProblem class instance.
        initial_guess (Optional[np.ndarray]): The initial guess of decision variables
            to be used in the solver. If None, uses the lower bounds defined for
            the variables in MOProblem. Defaults to None.
        solver_method (Optional[Union[ScalarMethod, str]]): The method used to minimize the
            invidual problems in the payoff table method. Defaults to 'scipy_de'.
    
    Returns:
        Tuple[np.ndarray, np.ndarray]: The ideal and nadir points
    """
    if problem.n_of_constraints > 0:
        constraints = lambda x: problem.evaluate(x).constraints.squeeze()
    else:
        constraints = None

    return payoff_table_method_general(
        lambda xs: problem.evaluate(xs).objectives,
        problem.n_of_objectives,
        problem.get_variable_bounds(),
        constraints,
        initial_guess,
        solver_method,
    )


def solve_pareto_front_representation_general(
    objective_evaluator: Callable[[np.ndarray], np.ndarray],
    n_of_objectives: int,
    variable_bounds: np.ndarray,
    step: Optional[Union[np.ndarray, float]] = 0.1,
    eps: Optional[float] = 1e-6,
    ideal: Optional[np.ndarray] = None,
    nadir: Optional[np.ndarray] = None,
    constraint_evaluator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    solver_method: Optional[Union[ScalarMethod, str]] = "scipy_de",
) -> Tuple[np.ndarray, np.ndarray]:
    """Computes a representation of a Pareto efficient front from a
    multiobjective minimizatino problem. Does so by generating an evenly spaced
    set of reference points (in the objective space), in the space spanned by
    the supplied ideal and nadir points. The generated reference points are
    then used to formulate achievement scalaraization problems, which when
    solved, yield a representation of a Pareto efficient solution. The result
    is guaranteed to contain only non-dominated solutions.
    
    Args:
        objective_evaluator (Callable[[np.ndarray], np.ndarray]): A vector
            valued function returning objective values given an array of decision
            variables.
        n_of_objectives (int): Numbr of objectives returned by
            objective_evaluator.
        variable_bounds (np.ndarray): The upper and lower bounds of the
            decision variables. Bound for each variable should be on the rows,
            with the first column containing lower bounds, and the second column
            upper bounds. Use np.inf to indicate no bounds.
        step (Optional[Union[np.ndarray, float]], optional): Etiher an float
            or an array of floats. If a single float is given, generates
            reference points with the objectives having values a step apart
            between the ideal and nadir points. If an array of floats is given,
            use the steps defined in the array for each objective's values.
            Default to 0.1.
        eps (Optional[float], optional): An offset to be added to the nadir
            value to keep the nadir inside the range when generating reference
            points. Defaults to 1e-6.
        ideal (Optional[np.ndarray], optional): The ideal point of the
            problem being solved. Defaults to None.
        nadir (Optional[np.ndarray], optional): The nadir point of the
            problem being solved. Defaults to None.
        constraint_evaluator (Optional[Callable[[np.ndarray], np.ndarray]], optional):
            An evaluator returning values for the constraints defined
            for the problem. A negative value for a constraint indicates a breach
            of that constraint. Defaults to None.
        solver_method (Optional[Union[ScalarMethod, str]], optional): The
            method used to minimize the achievement scalarization problems
            arising when calculating Pareto efficient solutions. Defaults to
            "scipy_de".

    Raises:
        MCDMUtilityException: Mismatching sizes of the supplied ideal and
        nadir points between the step, when step is an array. Or the type of
        step is something else than np.ndarray of float.
    
    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple containing representationns of
        the Pareto optimal variable values, and the corresponsing objective
        values.
    """
    if ideal is None or nadir is None:
        # compure ideal and nadir using payoff table
        ideal, nadir = payoff_table_method_general(
            objective_evaluator, n_of_objectives, variable_bounds, constraint_evaluator,
        )

    # use ASF to (almost) gurantee Pareto optimality.
    asf = PointMethodASF(nadir, ideal)

    scalarizer = Scalarizer(objective_evaluator, asf, scalarizer_args={"reference_point": None})
    solver = ScalarMinimizer(scalarizer, bounds=variable_bounds, method=solver_method)

    if type(step) is float:
        slices = [slice(start, stop + eps, step) for (start, stop) in zip(ideal, nadir)]

    elif type(step) is np.ndarray:
        if not ideal.shape == nadir.shape == step.shape:
            raise MCDMUtilityException(
                "The shapes of the supplied step array does not match the " "shape of the ideal and nadir points."
            )
        slices = [slice(start, stop + eps, s) for (start, stop, s) in zip(ideal, nadir, step)]

    else:
        raise MCDMUtilityException("step must be either a numpy array or an float.")

    z_mesh = np.mgrid[slices].reshape(len(ideal), -1).T

    p_front_objectives = np.zeros(z_mesh.shape)
    p_front_variables = np.zeros((len(p_front_objectives), len(variable_bounds.squeeze())))

    for i, z in enumerate(z_mesh):
        scalarizer._scalarizer_args = {"reference_point": z}
        res = solver.minimize(None)

        if not res["success"]:
            print("Non successfull optimization")
            p_front_objectives[i] = np.nan
            p_front_variables[i] = np.nan
            continue

        # check for dominance, accept only non-dominated solutions
        f_i = objective_evaluator(res["x"])
        if not np.all(f_i > p_front_objectives[:i][~np.all(np.isnan(p_front_objectives[:i]), axis=1)]):
            p_front_objectives[i] = f_i
            p_front_variables[i] = res["x"]
        elif i < 1:
            p_front_objectives[i] = f_i
            p_front_variables[i] = res["x"]
        else:
            print(f"{f_i} is dominated by {p_front_objectives[:i][np.all(f_i > p_front_objectives[:i], axis=1)]}")
            p_front_objectives[i] = np.nan
            p_front_variables[i] = np.nan

    return (
        p_front_variables[~np.all(np.isnan(p_front_variables), axis=1)],
        p_front_objectives[~np.all(np.isnan(p_front_objectives), axis=1)],
    )


def solve_pareto_front_representation(
    problem: MOProblem,
    step: Optional[Union[np.ndarray, float]] = 0.1,
    eps: Optional[float] = 1e-6,
    solver_method: Optional[Union[ScalarMethod, str]] = "scipy_de",
) -> Tuple[np.ndarray, np.ndarray]:
    """Pass through to solve_pareto_front_representation_general when the
    problem for which the front is being calculated for is defined as an
    MOProblem object.
    
    Computes a representation of a Pareto efficient front
    from a multiobjective minimizatino problem. Does so by generating an
    evenly spaced set of reference points (in the objective space), in the
    space spanned by the supplied ideal and nadir points. The generated
    reference points are then used to formulate achievement scalaraization
    problems, which when solved, yield a representation of a Pareto efficient
    solution.
    
    Args:
        problem (MOProblem): The multiobjective minimization problem for which the front is to be solved for.
            step (Optional[Union[np.ndarray, float]], optional): Etiher an float
            or an array of floats. If a single float is given, generates
            reference points with the objectives having values a step apart
            between the ideal and nadir points. If an array of floats is given,
            use the steps defined in the array for each objective's values.
            Default to 0.1.
        eps (Optional[float], optional): An offset to be added to the nadir
            value to keep the nadir inside the range when generating reference
            points. Defaults to 1e-6.
        solver_method (Optional[Union[ScalarMethod, str]], optional): The
            method used to minimize the achievement scalarization problems
            arising when calculating Pareto efficient solutions. Defaults to
            "scipy_de".
    
    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple containing representationns of
        the Pareto optimal variable values, and the corresponsing objective
        values.
    """
    if problem.n_of_constraints > 0:
        constraints = lambda x: problem.evaluate(x).constraints.squeeze()
    else:
        constraints = None

    return solve_pareto_front_representation_general(
        lambda x: problem.evaluate(x).objectives,
        problem.n_of_objectives,
        problem.get_variable_bounds(),
        step,
        eps,
        problem.ideal,
        problem.nadir,
        constraints,
        solver_method,
    )


if __name__ == "__main__":
    from desdeo_problem.Problem import MOProblem
    from desdeo_problem.testproblems.TestProblems import test_problem_builder
    from desdeo_problem.Objective import _ScalarObjective
    from desdeo_problem.Variable import variable_builder
    from desdeo_problem.Constraint import ScalarConstraint
    from desdeo_tools.scalarization.Scalarizer import Scalarizer
    from desdeo_tools.solver.ScalarSolver import ScalarMethod

    from scipy.optimize import differential_evolution

    # # create the problem
    # def f_1(x):
    #     res = 4.07 + 2.27 * x[:, 0]
    #     return -res

    # def f_2(x):
    #     res = (
    #         2.60
    #         + 0.03 * x[:, 0]
    #         + 0.02 * x[:, 1]
    #         + 0.01 / (1.39 - x[:, 0] ** 2)
    #         + 0.30 / (1.39 - x[:, 1] ** 2)
    #     )
    #     return -res

    # def f_3(x):
    #     res = 8.21 - 0.71 / (1.09 - x[:, 0] ** 2)
    #     return -res

    # def f_4(x):
    #     res = 0.96 - 0.96 / (1.09 - x[:, 1] ** 2)
    #     return -res

    # def f_5(x):
    #     return np.max([np.abs(x[:, 0] - 0.65), np.abs(x[:, 1] - 0.65)], axis=0)

    # def c_1(x, f=None):
    #     x = x.squeeze()
    #     return (x[0] + x[1]) - 0.2

    # f1 = _ScalarObjective(name="f1", evaluator=f_1)
    # f2 = _ScalarObjective(name="f2", evaluator=f_2)
    # f3 = _ScalarObjective(name="f3", evaluator=f_3)
    # f4 = _ScalarObjective(name="f4", evaluator=f_4)
    # f5 = _ScalarObjective(name="f5", evaluator=f_5)
    # c1 = ScalarConstraint("c1", 2, 5, evaluator=c_1)
    # varsl = variable_builder(
    #     ["x_1", "x_2"],
    #     initial_values=[0.5, 0.5],
    #     lower_bounds=[0.3, 0.3],
    #     upper_bounds=[1.0, 1.0],
    # )
    # problem = MOProblem(variables=varsl, objectives=[f1, f2, f3, f4, f5],)

    scalar_method = ScalarMethod(
        lambda x, _, **ys: differential_evolution(x, **ys),
        method_args={"polish": True, "maxiter": 1000},
        use_scipy=True,
    )

    problem = test_problem_builder("DTLZ1", 3, 2)
    print(problem.get_variable_bounds())
    problem.ideal = np.zeros(2)
    problem.nadir = np.ones(2)

    res = solve_pareto_front_representation_general(
        lambda x: problem.evaluate(x).objectives,
        problem.n_of_objectives,
        problem.get_variable_bounds(),
        step=0.2,
        eps=1e-6,
        ideal=problem.ideal,
        nadir=problem.nadir,
        constraint_evaluator=None,
        solver_method=scalar_method,
    )

    print(res[1].shape)



from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from desdeo_tools.interaction.request import BaseRequest
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin_min

from desdeo_mcdm.interactive.InteractiveMethod import InteractiveMethod


class ENautilusException(Exception):
    """Raised when an exception related to ENautilus is encountered.

    """

    pass


class ENautilusInitialRequest(BaseRequest):
    """ A request class to handle the initial preferences.

    """

    def __init__(self, ideal: np.ndarray, nadir: np.ndarray):
        msg = (
            "Please specify the number of iterations as 'n_iterations' to be carried out, and how many intermediate "
            "points to show as 'n_points'."
        )
        content = {
            "message": msg,
            "ideal": ideal,
            "nadir": nadir,
        }

        super().__init__("reference_point_preference", "required", content=content)

    def validator(self, response: Dict) -> None:
        if "n_iterations" not in response:
            raise ENautilusException("'n_iterations' entry missing")

        if "n_points" not in response:
            raise ENautilusException("'n_points' entry missing")

        n_iterations = response["n_iterations"]

        if not isinstance(n_iterations, int) or int(n_iterations) < 1:
            raise ENautilusException("'n_iterations' must be a positive integer greater than zero")

        n_points = response["n_points"]

        if not isinstance(n_points, int) or int(n_points) < 1:
            raise ENautilusException("'n_points' must be a positive integer greater than zero")

    @classmethod
    def init_with_method(cls, method):
        return cls(method._ideal, method._nadir)

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class ENautilusRequest(BaseRequest):
    """A request class to handle the intermediate requests.

    """

    def __init__(
        self,
        ideal: np.ndarray,
        nadir: np.ndarray,
        points: np.ndarray,
        lower_bounds: np.ndarray,
        upper_bounds: np.ndarray,
        distances: np.ndarray,
        minimize: List[int],
    ):
        self._max_index = len(np.squeeze(points))
        msg = "Please select the most preferred point by index as 'preferred_point_index'"
        content = {
            "message": msg,
            "ideal": ideal,
            "nadir": nadir,
            "points": points,
            "lower_bounds": lower_bounds,
            "upper_bounds": upper_bounds,
            "distances": distances,
            "minimize": minimize,
        }

        super().__init__("reference_point_preference", "required", content=content)

    def validator(self, response: Dict) -> None:
        if "preferred_point_index" not in response:
            raise ENautilusException("'preferred_point' not specified")

        pref_point_index = response["preferred_point_index"]

        if pref_point_index < 0 or pref_point_index > self._max_index:
            raise ENautilusException("The given index is out of bounds.")

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class ENautilusStopRequest(BaseRequest):
    """A request class to handle termination.

    """

    def __init__(self, preferred_point: np.ndarray):
        msg = "Most preferred solution found."
        content = {"message": msg, "solution": preferred_point}

        super().__init__("print", "no_interaction", content=content)


class ENautilus(InteractiveMethod):
    def __init__(
        self,
        pareto_front: np.ndarray,
        ideal: np.ndarray,
        nadir: np.ndarray,
        objective_names: Optional[List[str]] = None,
        minimize: Optional[List[int]] = None,
    ):
        """

        Args:
            pareto_front (np.ndarray): A two dimensional numpy array
            representing a Pareto front with objective vectors on each of its
            rows.
            ideal (np.ndarray): The ideal objective vector of the problem
            being represented by the Pareto front.
            nadir (np.ndarray): The nadir objective vector of the problem
            being represented by the Pareto front.
            objective_names (Optional[List[str]], optional): Names of the
            objectives. List must match the number of columns in
            pareto_front. Defaults to 'f1', 'f2', 'f3', ...
            minimize (Optional[List[int]], optional): Multipliers for each
            objective. '-1' indicates maximization and '1' minimization.
            Defaults to all objective values being minimized.

        Raises:
            ENavigatorException: One or more dimension mismatches are
            encountered among the supplies arguments.
        """
        if not pareto_front.ndim == 2:
            raise ENautilusException(
                "The supplied Pareto front should be a two dimensional array. Found "
                f" number of dimensions {pareto_front.ndim}."
            )

        if not ideal.shape[0] == pareto_front.shape[1]:
            raise ENautilusException(
                "The Pareto front must consist of objective vectors with the "
                "same number of objectives as defined in the ideal and nadir "
                "points."
            )

        if not ideal.shape == nadir.shape:
            raise ENautilusException("The dimensions of the ideal and nadir point do not match.")

        if objective_names:
            if not len(objective_names) == ideal.shape[0]:
                raise ENautilusException(
                    "The supplied objective names must have a leangth equal to " "the numbr of objectives."
                )
            self._objective_names = objective_names
        else:
            self._objective_names = [f"f{i+1}" for i in range(ideal.shape[0])]

        if minimize:
            if not len(objective_names) == ideal.shape[0]:
                raise ENautilusException("The minimize list must have " "as many elements as there are objectives.")
            self._minimize = minimize
        else:
            self._minimize = [1 for _ in range(ideal.shape[0])]

        self._ideal = ideal
        self._nadir = nadir

        # in objective space!
        self._pareto_front = pareto_front

        # bounds of the rechable region
        self._reachable_ub = self._nadir
        self._reachable_lb = self._ideal

        # currently reachable solution as a list of indices of the Pareto front
        self._reachable_idx = list(range(0, self._pareto_front.shape[0]))

        # current iteration step number
        self._step_number = 1

        self._distance = None

        self._preferred_point = None
        self._projection_index = None

        self._n_iterations = None
        self._n_points = None
        self._n_iterations_left = None

    def start(self) -> ENautilusInitialRequest:
        return ENautilusInitialRequest.init_with_method(self)

    def iterate(
        self, request: Union[ENautilusInitialRequest, ENautilusRequest]
    ) -> Union[ENautilusRequest, ENautilusStopRequest]:
        """Perform the next logical iteratino step based on the given request type.

        """
        if type(request) is ENautilusInitialRequest:
            return self.handle_initial_request(request)
        elif type(request) is ENautilusRequest:
            return self.handle_request(request)
        else:
            # if stop request, do nothing
            return request

    def handle_initial_request(self, request: ENautilusInitialRequest) -> ENautilusRequest:
        """Handles the initial request by parsing the response appropiately.

        """
        self._n_iterations = request.response["n_iterations"]
        self._n_points = request.response["n_points"]
        self._n_iterations_left = self._n_iterations
        # self._intermediate_points = np.repeat(np.atleast_2d(self._nadir), self._n_points, axis=0)
        self._preferred_point = self._nadir

        zbars = self.calculate_representative_points(self._pareto_front, self._reachable_idx, self._n_points)
        zs = self.calculate_intermediate_points(self._preferred_point, zbars, self._n_iterations_left)
        new_lower_bounds, new_upper_bounds = self.calculate_bounds(self._pareto_front, zs)
        distances = self.calculate_distances(zs, zbars, self._nadir)

        return ENautilusRequest(
            self._ideal, self._nadir, zs, new_lower_bounds, new_upper_bounds, distances, self._minimize
        )

    def handle_request(self, request: ENautilusRequest) -> Union[ENautilusRequest, ENautilusStopRequest]:
        """Handles the intermediate requests.

        """
        preferred_point_index = request.response["preferred_point_index"]
        self._preferred_point = request.content["points"][preferred_point_index]

        if self._n_iterations_left <= 1:
            self._n_iterations_left = 0
            return ENautilusStopRequest(self._preferred_point)

        self._reachable_lb = request.content["lower_bounds"][preferred_point_index]
        self._reachable_ub = request.content["upper_bounds"][preferred_point_index]
        self._distance = request.content["distances"][preferred_point_index]

        self._reachable_idx = self.calculate_reachable_point_indices(
            self._pareto_front, self._reachable_lb, self._reachable_ub
        )

        # increment and decrement iterations
        self._n_iterations_left -= 1
        self._step_number += 1

        # Start again
        zbars = self.calculate_representative_points(self._pareto_front, self._reachable_idx, self._n_points)
        zs = self.calculate_intermediate_points(self._preferred_point, zbars, self._n_iterations_left)
        new_lower_bounds, new_upper_bounds = self.calculate_bounds(self._pareto_front, zs)
        distances = self.calculate_distances(zs, zbars, self._nadir)

        return ENautilusRequest(
            self._ideal, self._nadir, zs, new_lower_bounds, new_upper_bounds, distances, self._minimize
        )

    def calculate_representative_points(
        self, pareto_front: np.ndarray, subset_indices: List[int], n_points: int
    ) -> np.ndarray:
        """Calcualtes the most representative points on the Pareto front. The points are clustered using k-means.

        Args:
            pareto_front (np.ndarray): The Pareto front.
            subset_indices (List[int]): A list of indices representing the
            subset of the points on the Pareto front for which the
            representative points should be calculated.
            n_points (int): The number of representative points to be calculated.

        Returns:
            np.ndarray: A 2D array of the most representative points. If the
            subset of Pareto efficient points is less than n_points, returns
            the subset of the Pareto front.
        """
        if len(np.atleast_1d(subset_indices)) > n_points:
            kmeans = KMeans(n_clusters=n_points)
            kmeans.fit(pareto_front[subset_indices])

            closest, _ = pairwise_distances_argmin_min(kmeans.cluster_centers_, pareto_front[subset_indices])

            zbars = pareto_front[subset_indices][closest]

        else:
            zbars = pareto_front[subset_indices]

        return zbars

    def calculate_intermediate_points(
        self, preferred_point: np.ndarray, zbars: np.ndarray, n_iterations_left: int,
    ) -> np.ndarray:
        """Calcualtes the intermediate points between representative points an a preferred point.

        Args:
            preferred_point (np.ndarray): The preferred point, 1D array.
            zbars (np.ndarray): The representative points, 2D array.
            n_iterations_left (int): The number of iterations left.

        Returns:
            np.ndarray: The intermediate points as a 2D array.
        """
        zs = ((n_iterations_left - 1) / n_iterations_left) * preferred_point + (1 / n_iterations_left) * zbars
        return np.atleast_2d(zs)

    def calculate_bounds(
        self, pareto_front: np.ndarray, intermediate_points: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the new bounds of the reachable points on the Pareto
        optimal front from each of the intermediate points.

        Args:
            pareto_front (np.ndarray): The Pareto optimal front.
            intermediate_points (np.ndarray): The current intermedaite points as a 2D array.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The lower and upper bounds for each of the intermediate points.
        """
        _pareto_front = np.atleast_2d(pareto_front)
        n_points = np.atleast_2d(intermediate_points).shape[0]
        new_lower_bounds = np.zeros((n_points, _pareto_front.shape[1]))
        new_upper_bounds = np.zeros((n_points, _pareto_front.shape[1]))

        for i, point in enumerate(np.atleast_2d(intermediate_points)):
            # TODO: vectorize this loop
            for r in range(_pareto_front.shape[1]):
                mask = np.zeros(_pareto_front.shape[1], dtype=bool)
                mask[r] = True

                subject_to = _pareto_front[:, ~mask].reshape((_pareto_front.shape[0], _pareto_front.shape[1] - 1))

                con_mask = np.all(subject_to <= point[~mask], axis=1)

                min_val = np.min(_pareto_front[con_mask, mask])
                max_val = np.max(_pareto_front[con_mask, mask])

                new_lower_bounds[i, r] = min_val
                new_upper_bounds[i, r] = max_val

        return new_lower_bounds, new_upper_bounds

    def calculate_distances(self, intermediate_points: np.ndarray, zbars: np.ndarray, nadir: np.ndarray) -> np.ndarray:
        distances = np.linalg.norm(np.atleast_2d(intermediate_points) - nadir, axis=1) / np.linalg.norm(
            np.atleast_2d(zbars) - nadir, axis=1
        )
        """Calculates the distance to the Pareto front for each intermediate
        point given utilizing representative points representing the
        intermediate points.

        Args:
            intermediate_points (np.ndarray): The intermediate points, 2D array.
            zbars (np.ndarray): The representative points corresponding to the intermediate points, 2D array.
            nadir (np.ndarray): The nadir point, 1D array.

        Returns:
            np.ndarray: The distances calculated for each intermediate point to the Pareto front.
        """
        return distances * 100

    def calculate_reachable_point_indices(
        self, pareto_front: np.ndarray, lower_bounds: np.ndarray, upper_bounds: np.ndarray,
    ) -> List[int]:
        """Calculate the indices of the reachable Pareto optimal solutions
        based on lower and upper bounds.

        Returns:
            List[int]: List of the indices of the reachable solutions.
        """
        low_idx = np.all(pareto_front >= lower_bounds, axis=1)
        up_idx = np.all(pareto_front <= upper_bounds, axis=1)

        reachable_idx = np.argwhere(low_idx & up_idx).squeeze()

        return reachable_idx


if __name__ == "__main__":
    # front = np.array([[1, 2, 3], [2, 3, 4], [2, 2, 3], [3, 2, 1]], dtype=float)
    # ideal = np.zeros(3)
    # nadir = np.ones(3) * 5
    f1 = np.linspace(1, 100, 50)
    f2 = f1[::-1] ** 2

    front = np.stack((f1, f2)).T
    ideal = np.min(front, axis=0)
    nadir = np.max(front, axis=0)

    method = ENautilus((front), ideal, nadir)

    req = method.start()

    n_iterations = 11
    n_points = 4

    req.response = {
        "n_iterations": n_iterations,
        "n_points": n_points,
    }

    req = method.iterate(req)
    req.response = {"preferred_point_index": 0}

    while method._n_iterations_left > 1:
        print(method._n_iterations_left)
        req = method.iterate(req)
        print(req.content["points"])
        req.response = {"preferred_point_index": 0}

    print(method._n_iterations_left)
    req = method.iterate(req)
    print(method._n_iterations_left)
    print(method._distance)
    print(req.content["solution"])



"""This module contains interactive methods and related requests implemented as classes.
"""

__all__ = [
    "ENautilus",
    "ENautilusException",
    "ENautilusInitialRequest",
    "ENautilusRequest",
    "ENautilusStopRequest",
    "NautilusNavigator",
    "NautilusNavigatorException",
    "NautilusNavigatorRequest",
    "NIMBUS",
    "NimbusException",
    "NimbusClassificationRequest",
    "NimbusIntermediateSolutionsRequest",
    "NimbusMostPreferredRequest",
    "NimbusSaveRequest",
    "NimbusStopRequest",
]

from desdeo_mcdm.interactive.ENautilus import (
    ENautilus,
    ENautilusException,
    ENautilusInitialRequest,
    ENautilusRequest,
    ENautilusStopRequest,
)
from desdeo_mcdm.interactive.NautilusNavigator import (
    NautilusNavigator,
    NautilusNavigatorException,
    NautilusNavigatorRequest,
)
from desdeo_mcdm.interactive.NIMBUS import (
    NIMBUS,
    NimbusClassificationRequest,
    NimbusException,
    NimbusIntermediateSolutionsRequest,
    NimbusMostPreferredRequest,
    NimbusSaveRequest,
    NimbusStopRequest,
)



from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from desdeo_mcdm.interactive.InteractiveMethod import InteractiveMethod
from desdeo_tools.interaction.request import BaseRequest, SimplePlotRequest
from desdeo_tools.scalarization.ASF import PointMethodASF
from desdeo_tools.scalarization.Scalarizer import DiscreteScalarizer
from desdeo_tools.solver.ScalarSolver import DiscreteMinimizer


class NautilusNavigatorRequest(BaseRequest):
    """Request to handle interactions with NAUTILUS Navigator. See the
    NautilusNavigator class for further details.
    """

    def __init__(
        self,
        ideal: np.ndarray,
        nadir: np.ndarray,
        reachable_lb: np.ndarray,
        reachable_ub: np.ndarray,
        reachable_idx: List[int],
        step_number: int,
        steps_remaining: int,
        distance: float,
        allowed_speeds: [int],
        current_speed: int,
        navigation_point: np.ndarray,
    ):
        msg = (
            # TODO: Be more specific...
            "Please supply aspirations levels for each objective between "
            "the upper and lower bounds as `reference_point`. Specify a "
            "speed between 1-5 as `speed`. If going to a previous step is "
            "desired, please set `go_to_previous` to True, otherwise it should "
            "be False. "
            "Lastly, if stopping is desired, `stop` should be True, "
            "otherweise it should be set to False."
        )
        content = {
            "message": msg,
            "ideal": ideal,
            "nadir": nadir,
            "reachable_lb": reachable_lb,
            "reachable_ub": reachable_ub,
            "reachable_idx": reachable_idx,
            "step_number": step_number,
            "steps_remaining": steps_remaining,
            "distance": distance,
            "allowed_speeds": allowed_speeds,
            "current_speed": current_speed,
            "navigation_point": navigation_point,
        }

        super().__init__(
            "reference_point_preference", "required", content=content
        )

    @classmethod
    def init_with_method(cls, method):
        return cls(
            method._ideal,
            method._nadir,
            method._reachable_lb,
            method._reachable_ub,
            method._reachable_idx,
            method._step_number,
            method._steps_remaining,
            method._distance,
            method._allowed_speeds,
            method._current_speed,
            method._navigation_point,
        )

    def validator(self, response: Dict) -> None:
        if "reference_point" not in response:
            raise NautilusNavigatorException("'reference_point' entry missing.")

        if "speed" not in response:
            raise NautilusNavigatorException("'speed' entry missing.")

        if "go_to_previous" not in response:
            raise NautilusNavigatorException("'go_to_previous' entry missing.")

        if "stop" not in response:
            raise NautilusNavigatorException("'stop' entry missing.")

        ref_point = response["reference_point"]
        try:
            if np.any(ref_point < self._content["ideal"]) or np.any(
                ref_point > self._content["nadir"]
            ):
                raise NautilusNavigatorException(
                    f"The given reference point {ref_point} "
                    "must be between the ranges imposed by the ideal and nadir points."
                )
        except Exception as e:
            raise NautilusNavigatorException(
                f"An exception rose when validating the given reference point {ref_point}.\n"
                f"Previous exception: {type(e)}: {str(e)}."
            )

        speed = response["speed"]
        try:
            if int(speed) not in self._content["allowed_speeds"]:
                raise NautilusNavigatorException(f"Invalid speed: {speed}.")
        except Exception as e:
            raise NautilusNavigatorException(
                f"An exception rose when validating the given speed {speed}.\n"
                f"Previous exception: {type(e)}: {str(e)}."
            )

        if not type(response["go_to_previous"]) == bool:
            raise (
                f"Non boolean value {response['go_to_previous']} "
                f"found for 'go_to_previous' when validating the response."
            )

        if not type(response["stop"]) == bool:
            raise (
                f"Non boolean value {response['stop']} "
                f"found for 'go_to_previous' when validating the response."
            )

    @BaseRequest.response.setter
    def response(self, response: Dict):
        self.validator(response)
        self._response = response


class NautilusNavigatorException(Exception):
    """Raised when an exception related to NAUTILUS Navigator is encountered.
    """

    pass


class NautilusNavigator(InteractiveMethod):
    """Implementations of the NAUTILUS Navigator algorithm.

    """

    def __init__(
        self,
        pareto_front: np.ndarray,
        ideal: np.ndarray,
        nadir: np.ndarray,
        objective_names: Optional[List[str]] = None,
        minimize: Optional[List[int]] = None,
    ):
        """

        Args:
            pareto_front (np.ndarray): A two dimensional numpy array
            representing a Pareto front with objective vectors on each of its
            rows.
            ideal (np.ndarray): The ideal objective vector of the problem
            being represented by the Pareto front.
            nadir (np.ndarray): The nadir objective vector of the problem
            being represented by the Pareto front.
            objective_names (Optional[List[str]], optional): Names of the
            objectives. List must match the number of columns in
            pareto_front. Defaults to 'f1', 'f2', 'f3', ...
            minimize (Optional[List[int]], optional): Multipliers for each
            objective. '-1' indicates maximization and '1' minimization.
            Defaults to all objective values being minimized.

        Raises:
            NautilusNavigatorException: One or more dimension mismatches are
            encountered among the supplies arguments.
        """
        if not pareto_front.ndim == 2:
            raise NautilusNavigatorException(
                "The supplied Pareto front should be a two dimensional array. Found "
                f" number of dimensions {pareto_front.ndim}."
            )

        if not ideal.shape[0] == pareto_front.shape[1]:
            raise NautilusNavigatorException(
                "The Pareto front must consist of objective vectors with the "
                "same number of objectives as defined in the ideal and nadir "
                "points."
            )

        if not ideal.shape == nadir.shape:
            raise NautilusNavigatorException(
                "The dimensions of the ideal and nadir point do not match."
            )

        if objective_names:
            if not len(objective_names) == ideal.shape[0]:
                raise NautilusNavigatorException(
                    "The supplied objective names must have a leangth equal to "
                    "the numbr of objectives."
                )
            self._objective_names = objective_names
        else:
            self._objective_names = [f"f{i+1}" for i in range(ideal.shape[0])]

        if minimize:
            if not len(objective_names) == ideal.shape[0]:
                raise NautilusNavigatorException(
                    "The minimize list must have "
                    "as many elements as there are objectives."
                )
            self._minimize = minimize
        else:
            self._minimize = [1 for _ in range(ideal.shape[0])]

        self._ideal = ideal
        self._nadir = nadir

        # in objective space!
        self._pareto_front = pareto_front

        # bounds of the rechable region
        self._reachable_ub = self._nadir
        self._reachable_lb = self._ideal

        # currently reachable solution as a list of indices of the Pareto front
        self._reachable_idx = list(range(0, self._pareto_front.shape[0]))

        # current iteration step number
        self._step_number = 1

        # iterations left
        self._steps_remaining = 100

        # L2 distance to the supplied Pareto front
        self._distance = 0

        self._allowed_speeds = [1, 2, 3, 4, 5]
        self._current_speed = None

        self._reference_point = None
        self._navigation_point = self._nadir
        self._projection_index = None

    def start(self) -> NautilusNavigatorRequest:
        """Returns the first Request object to begin iterating.

        Returns:
            NautilusNavigatorRequest: The Request.
        """
        return NautilusNavigatorRequest.init_with_method(self)

    def iterate(
        self, request: NautilusNavigatorRequest
    ) -> NautilusNavigatorRequest:
        """Perform the next logical step based on the response in the
        Request.
        """
        reqs = self.handle_request(request)
        return reqs

    def handle_request(
        self, request: NautilusNavigatorRequest
    ) -> NautilusNavigatorRequest:
        """Handle the Request and its contents.

        Args:
            request (NautilusNavigatorRequest): A Request with a defined response.

        Returns:
            NautilusNavigatorRequest: Some of the contents of the response are invalid.
        """
        preference_point = request.response["reference_point"]
        speed = request.response["speed"]
        go_to_previous = request.response["go_to_previous"]
        stop = request.response["stop"]

        if go_to_previous:
            step_number = request.content["step_number"]
            nav_point = request.content["navigation_point"]
            lower_bounds = request.content["reachable_lb"]
            upper_bounds = request.content["reachable_ub"]
            reachable_idx = request.content["reachable_idx"]
            distance = request.content["distance"]
            steps_remaining = request.content["steps_remaining"]

            return self.update(
                preference_point,
                speed,
                go_to_previous,
                stop,
                step_number,
                nav_point,
                lower_bounds,
                upper_bounds,
                reachable_idx,
                distance,
                steps_remaining,
            )

        else:
            return self.update(preference_point, speed, go_to_previous, stop,)

    def update(
        self,
        ref_point: np.ndarray,
        speed: int,
        go_to_previous: bool,
        stop: bool,
        step_number: Optional[int] = None,
        nav_point: Optional[np.ndarray] = None,
        lower_bounds: Optional[np.ndarray] = None,
        upper_bounds: Optional[np.ndarray] = None,
        reachable_idx: Optional[List[int]] = None,
        distance: Optional[float] = None,
        steps_remaining: Optional[int] = None,
    ) -> Optional[NautilusNavigatorRequest]:
        """Update the inernal state of self.

        Args:
            ref_point (np.ndarray): A reference point given by a decision maker.
            speed (int): An integer value between 1-5 indicating the navigation speed.
            go_to_previous (bool): If True, the parameters indicate the state
            of a previous state, and the request is handled accordingly.
            stop (bool): If the navigation should stop. If True, self.update return None.
            step_number (Optional[int], optional): Current step number, or
            previous step number if go_to_previous is True. Defaults to None.
            nav_point (Optional[np.ndarray], optional): The current
            navigation point. Relevant if go_to_previous is True. Defaults to
            None.
            lower_bounds (Optional[np.ndarray], optional): Lower bounds of
            the reachable objective vector valus. Relevant if go_to_previous
            is True. Defaults to None.
            upper_bounds (Optional[np.ndarray], optional): Upper bounds of
            the reachable objective vector valus. Relevant if go_to_previous
            is True. Defaults to None.
            reachable_idx (Optional[List[int]], optional): Indices of the
            reachable Pareto optimal solutions. Relevant if go_to_previous is
            True. Defaults to None.
            distance (Optional[float], optional): Distance to the Pareto
            optimal front. Relevant if go_to_previous is True. Defaults to
            None.
            steps_remaining (Optional[int], optional): Remaining steps in the
            navigation. Relevant if go_to_previous is True. Defaults to None.

        Returns:
            NautilusNavigatorRequest: Some of the given parameters are erraneous.
        """
        # go to a previous state
        if go_to_previous:
            self._step_number = step_number
            self._navigation_point = nav_point
            self._reachable_lb = lower_bounds
            self._reachable_ub = upper_bounds
            self._reachable_idx = reachable_idx
            self._distance = distance
            self._steps_remaining = steps_remaining
            return NautilusNavigatorRequest.init_with_method(self)

        # compute a new navigation point closer to the Pareto front and the
        # bounds of the reachable Pareto optimal region.
        elif self._step_number == 1 or not np.allclose(
            ref_point, self._reference_point
        ):
            if self._step_number == 1:
                self._current_speed = speed

            proj_i = self.solve_nautilus_asf_problem(
                self._pareto_front,
                self._reachable_idx,
                ref_point,
                self._ideal,
                self._nadir,
            )

            self._reference_point = ref_point
            self._projection_index = proj_i

        elif stop:
            # new reference point given, also update speed
            self._reference_point = ref_point
            self._current_speed = speed

        new_nav = self.calculate_navigation_point(
            self._pareto_front[self._projection_index],
            self._navigation_point,
            self._steps_remaining,
        )

        self._navigation_point = new_nav

        new_lb, new_ub = self.calculate_bounds(
            self._pareto_front[self._reachable_idx], self._navigation_point,
        )

        self._reachable_lb = new_lb
        self._reachable_ub = new_ub

        new_dist = self.calculate_distance(
            self._navigation_point,
            self._pareto_front[self._projection_index],
            self._nadir,
        )

        self._distance = new_dist

        new_reachable = self.calculate_reachable_point_indices(
            self._pareto_front, self._reachable_lb, self._reachable_ub,
        )

        self._reachable_idx = new_reachable

        # If stop, do not update steps
        if self._steps_remaining == 1:
            # stop
            return NautilusNavigatorRequest.init_with_method(self)

        self._step_number += 1
        self._steps_remaining -= 1

        return NautilusNavigatorRequest.init_with_method(self)

    def calculate_reachable_point_indices(
        self,
        pareto_front: np.ndarray,
        lower_bounds: np.ndarray,
        upper_bounds: np.ndarray,
    ) -> List[int]:
        """Calculate the indices of the reachable Pareto optimal solutions
        based on lower and upper bounds.

        Returns:
            List[int]: List of the indices of the reachable solutions.
        """
        low_idx = np.all(pareto_front >= lower_bounds, axis=1)
        up_idx = np.all(pareto_front <= upper_bounds, axis=1)

        reachable_idx = np.argwhere(low_idx & up_idx).squeeze()

        return reachable_idx

    def solve_nautilus_asf_problem(
        self,
        pareto_f: np.ndarray,
        subset_indices: [int],
        ref_point: np.ndarray,
        ideal: np.ndarray,
        nadir: np.ndarray,
    ) -> int:
        """Forms and solves the achievement scalarizing function to find the
        closesto point on the Pareto optimal front to the given reference
        point.

        Args:
            pareto_f (np.ndarray): The whole Pareto optimal front.
            subset_indices ([type]): Indices of the currently reachable solutions.
            ref_point (np.ndarray): The reference point indicating a decision
            maker's preference.
            ideal (np.ndarray): Ideal point.
            nadir (np.ndarray): Nadir point.

        Returns:
            int: Index of the closest point according the minimized value of the ASF.
        """
        asf = PointMethodASF(nadir, ideal)
        scalarizer = DiscreteScalarizer(asf, {"reference_point": ref_point})
        solver = DiscreteMinimizer(scalarizer)

        tmp = np.copy(pareto_f)
        mask = np.zeros(tmp.shape[0], dtype=bool)
        mask[subset_indices] = True
        tmp[~mask] = np.nan

        res = solver.minimize(tmp)

        return res

    def calculate_navigation_point(
        self,
        projection: np.ndarray,
        nav_point: np.ndarray,
        steps_remaining: int,
    ) -> np.ndarray:
        """Calculate a new navigation point based on the projection of the
        preference point to the Pareto optimal front.

        Args:
            projection (np.ndarray): The point on the Pareto optimal front
            closest to the preference point given by a decision maker.
            nav_point (np.ndarray): The previous navigation point.
            steps_remaining (int): How many steps are remaining in the navigation.

        Returns:
            np.ndarray: The new navigation point.
        """
        new_nav_point = (
            (steps_remaining - 1) / steps_remaining
        ) * nav_point + (1 / steps_remaining) * projection
        return new_nav_point

    def calculate_bounds(
        self, pareto_front: np.ndarray, nav_point: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the new bounds of the reachable points on the Pareto
        optimal front from a navigation point.

        Args:
            pareto_front (np.ndarray): The Pareto optimal front.
            nav_point (np.ndarray): The current navigation point.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The lower and upper bounds.
        """
        _pareto_front = np.atleast_2d(pareto_front)
        new_lower_bounds = np.zeros(_pareto_front.shape[1])
        new_upper_bounds = np.zeros(_pareto_front.shape[1])

        # TODO: vectorize this loop
        for r in range(_pareto_front.shape[1]):
            mask = np.zeros(_pareto_front.shape[1], dtype=bool)
            mask[r] = True

            subject_to = _pareto_front[:, ~mask].reshape(
                (_pareto_front.shape[0], _pareto_front.shape[1] - 1)
            )

            con_mask = np.all(subject_to <= nav_point[~mask], axis=1)

            min_val = np.min(_pareto_front[con_mask, mask])
            max_val = np.max(_pareto_front[con_mask, mask])

            new_lower_bounds[r] = min_val
            new_upper_bounds[r] = max_val

        return new_lower_bounds, new_upper_bounds

    def calculate_distance(
        self, nav_point: np.ndarray, projection: np.ndarray, nadir: np.ndarray
    ) -> float:
        """Calculate the distance to the Pareto optimal front from a
        navigation point. The distance is calculated to the supplied
        projection which is assumed to lay on the front.

        Args:
            nav_point (np.ndarray): The navigation point.
            projection (np.ndarray): The point of the Pareto optimal front the distance is calculated to.
            nadir (np.ndarray): The nadir point of the Pareto optimal set.

        Returns:
            float: The distance.
        """
        nom = np.linalg.norm(nav_point - nadir)
        denom = np.linalg.norm(projection - nadir)
        dist = (nom / denom) * 100

        return dist


if __name__ == "__main__":
    # front = np.array([[1, 2, 3], [2, 3, 4], [2, 2, 3], [3, 2, 1]], dtype=float)
    # ideal = np.zeros(3)
    # nadir = np.ones(3) * 5
    f1 = np.linspace(1, 100, 50)
    f2 = f1[::-1] ** 2

    front = np.stack((f1, f2)).T
    ideal = np.min(front, axis=0)
    nadir = np.max(front, axis=0)

    method = NautilusNavigator((front), ideal, nadir)

    req = method.start()
    print(req.content["reachable_lb"])
    print(req.content["navigation_point"])
    print(req.content["reachable_ub"])

    response = {
        "reference_point": np.array([50, 6000]),
        "speed": 5,
        "go_to_previous": False,
        "stop": False,
    }
    req.response = response
    req = method.iterate(req)
    req.response = response

    req1 = req

    import time

    while req.content["steps_remaining"] > 1:
        time.sleep(1 / req.content["current_speed"])
        req = method.iterate(req)
        req.response = response
        print(req.content["steps_remaining"])
        print(req.content["reachable_lb"])
        print(req.content["navigation_point"])
        print(req.content["reachable_ub"])

    req1.response["go_to_previous"] = True
    req = method.iterate(req1)
    req.response = response
    req.response["go_to_previous"] = False

    while req.content["steps_remaining"] > 1:
        time.sleep(1 / req.content["current_speed"])
        req = method.iterate(req)
        req.response = response
        print(req.content["steps_remaining"])
        print(req.content["reachable_lb"])
        print(req.content["navigation_point"])
        print(req.content["reachable_ub"])
    print(req)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['desdeo_mcdm', 'desdeo_mcdm.interactive', 'desdeo_mcdm.utilities']

package_data = \
{'': ['*']}

install_requires = \
['desdeo-problem>=0.15.0,<0.16.0', 'desdeo-tools>=0.2.6,<0.3.0']

setup_kwargs = {
    'name': 'desdeo-mcdm',
    'version': '0.1.2',
    'description': 'Contains traditional optimization techniques from the field of Multiple-criteria decision-making. Methods belonging to the NIMBUS and NAUTILUS families can be found here. Part of the DESDEO framework.',
    'long_description': None,
    'author': 'Giovanni Misitano',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from desdeo_problem.Problem import MOProblem


class InteractiveMethod:
    """The base class for interactive methods.

    Args:
        problem (MOProblem): The problem being solved in an interactive method.
    """

    def __init__(self, problem: MOProblem):
        self._problem = problem
